<?php
// This is called by cron
error_reporting(E_ALL);
ini_set("track_errors", "On");
set_time_limit(0);

$now = time();

require_once("/etc/rocketgit/config.php");

$INC = dirname(__FILE__) . "/../inc";
require_once($INC . "/init.inc.php");
require_once($INC . "/log.inc.php");
require_once($INC . "/sql.inc.php");
require_once($INC . "/struct.inc.php");
require_once($INC . "/repo.inc.php");
require_once($INC . "/keys.inc.php");
require_once($INC . "/fixes.inc.php");
require_once($INC . "/mr.inc.php");
require_once($INC . "/admin.inc.php");
require_once($INC . "/ver.php");

$user = isset($_SERVER['argv'][1]) ? $_SERVER['argv'][1] : "rocketgit";

if (strcmp($user, "apache") == 0) {
	rg_log_set_file($rg_web_log_dir . "/cron.log");
	rg_log_set_sid("000000"); // to spread the logs

	if (gmdate("Hi") == "0310")
		rg_clean_logs('/var/log/rocketgit-web');

	exit(0);
}

rg_log_set_file($rg_log_dir . "/cron.log");
rg_log_set_sid("000000"); // to spread the logs

if (rg_load() > 5)
	exit(0);

// locking
rg_lock_or_exit("cron.lock");

rg_log("Start (ver=$rocketgit_version)...");

if (gmdate("Hi") == "0305")
	rg_clean_logs('/var/log/rocketgit');

rg_sql_app("rg-cron");
$db = rg_sql_open($rg_sql);
if ($db === FALSE) {
	rg_log("Cannot connect to database (" . rg_sql_error() . ")!");
	// TODO: inform admin - already by e-mail?
	exit(1);
}

$restart_cache = FALSE;
if (rg_sql_struct_update_needed($db) == 1) {
	$r = rg_sql_struct_update($db, 0);
	if ($r !== TRUE)
		exit(1);
	$restart_cache = TRUE;
}

if (rg_fixes_needed($db) == 1) {
	$r = rg_fixes_update($db);
	if ($r !== TRUE)
		exit(1);
	$restart_cache = TRUE;
}

if ($restart_cache)
	rg_cache_restart();

$install_id = rg_state_get($db, "install_id");
if ($install_id === FALSE)
	exit(1);
if (empty($install_id))
	rg_state_set($db, "install_id", sha512(microtime(TRUE)));

$first_install = rg_state_get($db, 'first_install');
if ($first_install === FALSE)
	exit(1);
if (empty($first_install)) {
	rg_log("Admin did not created the user, so no need to run this code.");
	exit(0);
}

if (gmdate("Hi") == "0105") {
	while (1) {
		rg_log("Compute repository sizes if dirty...");
		// delete 'dirty' files
		$sql = "SELECT uid, repo_id, master, disk_used_mb FROM repos";
		$res = rg_sql_query($db, $sql);
		if ($res === FALSE) {
			// TODO: rg_internal_error? it must notify me in case of problems
			rg_log("Cannot run query (" . rg_sql_error() . ")!");
			break;
		}

		while (($row = rg_sql_fetch_array($res))) {
			rg_log("Processing repo " . $row['repo_id'] . "...");
			$repo_path = rg_repo_path_by_id($row['uid'], $row['repo_id']);

			$all_files = $row['master'] == 0 ? TRUE : FALSE;
			$disk_used = rg_repo_size($repo_path, $all_files);
			if ($disk_used === FALSE) {
				rg_log("Cannot compute the repo size: " . rg_repo_error());
				continue;
			}
			$disk_used_mb = sprintf("%u", $disk_used / 1024 / 1024);
			if ($disk_used_mb != $row['disk_used_mb']) {
				$sql = "UPDATE repos SET disk_used_mb = $disk_used_mb"
					. " WHERE repo_id = " . $row['repo_id'];
				$res2 = rg_sql_query($db, $sql);
				if ($res2 === FALSE) {
					rg_log("Cannot run query!");
					break;
				}
				rg_sql_free_result($res2);
			}
		}
		rg_sql_free_result($res);
		break;
	}

	while (1) {
		rg_log("Compute repository sizes per user...");
		$sql = "SELECT SUM(disk_used_mb) AS disk_used_mb, uid"
			. " FROM repos"
			. " GROUP BY uid";
		$res = rg_sql_query($db, $sql);
		if ($res === FALSE) {
			rg_log("Cannot run query (" . rg_sql_error() . ")!");
			break;
		}

		while (($row = rg_sql_fetch_array($res))) {
			$sql = "UPDATE users"
				. " SET disk_used_mb = " . $row['disk_used_mb']
				. " WHERE uid = " . $row['uid']
				. " AND disk_used_mb != " . $row['disk_used_mb'];
			$res2 = rg_sql_query($db, $sql);
			rg_sql_free_result($res2);
		}
		rg_sql_free_result($res);
		break;
	}
}

if (gmdate("Hi") == "0300") {
	rg_log("Clean old forget_pass entries...");
	$sql = "DELETE FROM forgot_pass WHERE expire < $now";
	$res = rg_sql_query($db, $sql);
	rg_sql_free_result($res);
}

if (gmdate("i") == "30") {
	rg_log("Clean old tokens...");
	$sql = "DELETE FROM tokens WHERE expire < $now";
	$res = rg_sql_query($db, $sql);
	rg_sql_free_result($res);
}

if (gmdate("i") == "01") {
	rg_log("Clean old sess entries...");
	$sql = "DELETE FROM sess WHERE expire < $now";
	$res = rg_sql_query($db, $sql);
	rg_sql_free_result($res);
}

if (gmdate("Hi") == "0605") {
	rg_log("Clean old login_tokens_ip entries...");
	$sql = "DELETE FROM login_tokens_ip WHERE expire < $now";
	$res = rg_sql_query($db, $sql);
	rg_sql_free_result($res);
}

rg_log_cron();

rg_sql_struct_slaves_update($db);

if (gmdate("Hi") == "0100")
	rg_admin_report1($db, $rg);

// TODO: move it as an event after the push
rg_mr_queue_process($db);

rg_log("Done!");
?>
